@@ -9,6 +9,16 @@ pub trait TypeInfo: Debug + Display + Clone + PartialEq<Self> + Send + Sync {
     /// should be a rough approximation of how they are written in SQL in the given database.
     fn name(&self) -> &str;
 
+    /// Return `true` if `self` and `other` represent mutually compatible types.
+    ///
+    /// Defaults to `self == other`.
+    fn type_compatible(&self, other: &Self) -> bool
+    where
+        Self: Sized,
+    {
+        self == other
+    }
+
     #[doc(hidden)]
     fn is_void(&self) -> bool {
         false
